/**
 * @author Bernhard Aichinger
 * matr#:  9825670
 *
 * @author Peter Praxmarer
 * matr#:  9755439
 *
 * Implements the logical-time mutual exclusion algorithm
 */

import java.util.LinkedList;

public class RicartAgrawalaME extends LogicalTimeME {

    // OK messages which are not sent now
    protected LinkedList[] deferred;

    // indicates if the OK messages have been sent
    protected boolean priorBcast;

    // indicates if the EXIT is reached through CRIT
    //protected boolean fromCrit;

    // last sent TRY messages
    protected Msg lastTry;



    public RicartAgrawalaME(int rank, int size) {
        super(rank, size);

        priorBcast = false;
        lastTry = null;
        //fromCrit = false;

        deferred = new LinkedList[size];
        for (int i = 0; i < size; i++)
            deferred[i] = new LinkedList();
    }



    protected void transitionTRY() {
        super.transitionTRY();
        lastTry = new Msg(Msg.TRY, clock, rank);
        priorBcast = true;
    }


    protected void transitionBCAST() {
        super.transitionBCAST();
        priorBcast = false;
    }


    protected void transitionRECEIVE() {
        int sender = in().select(1);
        while (sender != -1) {
            Msg msg = (Msg)in(sender).receive();
            if (msg != null) {
                int c = msg.getClock();
                int m = msg.getType();
                
                clock = (clock > c ? clock : c) + 1;
                history[sender].addLast(msg);
                if (m == Msg.TRY && sender != rank) {
                    Msg msgOk = new Msg(Msg.OK, clock, rank);

                    // point 1
                    // condition: R || E || T && prior broadcast
                    if (region == regionREM || region == regionEXIT ||
                        (region == regionTRY && priorBcast)) {
                        send[sender].addLast(msgOk);
                        System.out.println(rank + ": p1: ok message to " + sender);
                    }

                    // point 2
                    // condition: C
                    else if (region == regionCRIT) {
                        deferred[sender].addLast(msgOk);
                        System.out.println(rank + ": p2: deferred message to " + sender);
                    }

                    // point 3
                    // condition: T && after broadcast
                    else if (region == regionTRY && !priorBcast) {
                        if (lastTry.getClock() > msg.getClock() ||
                            (lastTry.getClock() == msg.getClock() && rank > sender)) {
                            send[sender].addLast(msgOk);
                            System.out.println(rank + ": p3: ok message to " + sender);
                        }
                        else {
                            deferred[sender].addLast(msgOk);
                            System.out.println(rank + ": p3: deferred message to " + sender);
                        }
                    }
		}
	    }
            sender = in().select(1);
        }
    }


    protected void transitionCRIT() {
        if (region == regionTRY) {
            boolean foundTry, foundOk;
            Msg msgTry = null;
	    
	    foundTry = lastTry != null;
	    msgTry = lastTry;
	 
            // count OK messages
            int counter = 0;
            for (int i = history.length - 1; i >= 0 && foundTry && msgTry != null; i--) {
                foundOk = false;
                for (int j = history[i].size() - 1; j >= 0 && !foundOk; j--) {
                    Msg msgOk = (Msg)history[i].get(j);
                    if (msgOk.getType() == Msg.OK &&
                        (msgOk.getClock() > msgTry.getClock() ||
                         msgOk.getClock() == msgTry.getClock() && i > rank)) {
                        foundOk = true;
                        counter++;
                    }
		}
            }
            foundOk = (counter == (history.length - 1));

            if (foundTry && foundOk) {
                clock++;
                region = regionCRIT;

		dumpState();

                interrupt();
            }
        }
    }


    protected void transitionEXIT() {
        super.transitionEXIT();

        // send deferred OK messages after C (from point 2)
        // send deferred OK messages after T -> C (from point 3)
        // deferred messages are also sent after T -> E !!!
        //if (fromCrit) {
        //    fromCrit = false;
            for (int i = 0; i < deferred.length; i++)
                for (int j = deferred[i].size() - 1; j >= 0; j--) {
                    send[i].addLast(deferred[i].removeFirst());
                    //System.out.println(rank + ": send deferred message to " + i);
                }
        //}
    }

}
